%{
(* This program is free software; you can redistribute it and/or modify it under  *)
(* the terms of the GNU General Public License as published by the Free Software  *)
(* Foundation; either version 2 of the License, or (at your option) any later     *)
(* version.                                                                       *)
(*                                                                                *)  
(* This program is distributed in the hope that it will be useful, but WITHOUT    *)
(* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS  *)
(* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. *)
(*                                                                                *)
(* Programmed by Alok Menghrajani.                                                *)
(* http://www.alokonline.com/projets/compiler/simpleb.html                        *)
(*                                                                                *)
(* Description: this file is used by ocamlyacc to create grammar.ml, which        *)
(*              contains the function "Program". Grammar.program takes a lexical  *)
(*              buffer (created by Lexing.from_channel) and constructs an         *)
(*              abstract syntax tree (Absyn.program). If the token stream fails   *)
(*              to match an expected case, parse_error is called and the          *)
(*              exception Parsing.Parse_error is raised.                          *)
(*                                                                                *)
(* X-Reference: absyn.ml, lexer.mll, simpleb.ml (the main program)                *)


module A = Absyn
module E = Errormsg

let parse_error s =
  E.error (symbol_end () + 1, symbol_end() + 1) s

let context l r :A.pos = (rhs_start l, rhs_end r - 1)

%}

%token <Symbol.symbol> ID
%token <int32> INTCONST
%token BREAK CONTINUE ELSE FALSE FOR IF INT TRUE WHILE WRITEINT
%token NOT AND OR XOR EQ NE LT LE GT GE COMMA
%token PLUSPLUS MINUSMINUS
%token LBRACE RBRACE LPAREN RPAREN SEMICOL PLUS MINUS TIMES DIVIDE ASSIGN EOF MOD NEG PLUSASSIGN MINUSASSIGN TIMESASSIGN DIVIDEASSIGN MODASSIGN
%start program
%type <Absyn.program> program


%left XOR
%left OR
%left AND
%left EQ NE
%left LT LE GT GE
%left PLUS MINUS
%left TIMES DIVIDE MOD
%nonassoc NEG
%nonassoc NOT

%%

program :
  LBRACE declseq stmtseq RBRACE EOF
  { A.Program(A.VarDecSeq $2, A.SeqStmt $3, E.currentContext()) }

declseq :
  /* empty */
  { [] }
| decl declseq
  { $1@$2 }

decl :
  INT identlist SEMICOL
  { $2 }

identlist :
  ID
  { [($1, A.IntExp Int32.zero, context 1 1)] }
| ID COMMA identlist
  { ($1, A.IntExp Int32.zero, context 1 1)::$3 }

stmtseq :
  /* empty */
  {[]}
| stmt stmtseq
  { ($1, context 1 1)::$2}

stmt :
  simps SEMICOL
  { A.SeqStmt ($1) }
| control
  { $1 }

simps :
  /* empty */
  { [] }
| simp csimps
  { [($1, context 1 1)]@$2 }

csimps :
  /* empty */
  { [] }
| COMMA simp csimps
  { ($2, context 2 2)::$3 }

simp :
  ID ASSIGN intexp
  { A.AssignStmt(A.SimpleVar($1, context 1 1), $3, context 1 3) }
| ID PLUSASSIGN intexp
  { A.AssignStmt(A.SimpleVar($1, context 1 1),
	A.OpExp(A.VarExp(A.SimpleVar($1, context 1 1)), 
		A.PlusOp, $3, context 3 3), context 1 3) }
| ID MINUSASSIGN intexp
  { A.AssignStmt(A.SimpleVar($1, context 1 1),
	A.OpExp(A.VarExp(A.SimpleVar($1, context 1 1)), 
		A.MinusOp, $3, context 3 3), context 1 3) }
| ID TIMESASSIGN intexp
  { A.AssignStmt(A.SimpleVar($1, context 1 1),
	A.OpExp(A.VarExp(A.SimpleVar($1, context 1 1)), 
		A.TimesOp, $3, context 3 3), context 1 3) }
| ID DIVIDEASSIGN intexp
  { A.AssignStmt(A.SimpleVar($1, context 1 1),
	A.OpExp(A.VarExp(A.SimpleVar($1, context 1 1)), 
		A.DivideOp, $3, context 3 3), context 1 3) }
| ID MODASSIGN intexp
  { A.AssignStmt(A.SimpleVar($1, context 1 1),
	A.OpExp(A.VarExp(A.SimpleVar($1, context 1 1)), 
		A.ModOp, $3, context 3 3), context 1 3) }
| PLUSPLUS ID
  { A.AssignStmt(A.SimpleVar($2, context 2 2),
        A.OpExp(A.VarExp(A.SimpleVar($2, context 2 2)),
                A.PlusOp, A.IntExp Int32.one, context 1 1), context 1 2) }
| ID PLUSPLUS
  { A.AssignStmt(A.SimpleVar($1, context 1 1),
        A.OpExp(A.VarExp(A.SimpleVar($1, context 1 1)),
                A.PlusOp, A.IntExp Int32.one, context 1 1), context 1 2) }
| MINUSMINUS ID
  { A.AssignStmt(A.SimpleVar($2, context 2 2),
        A.OpExp(A.VarExp(A.SimpleVar($2, context 2 2)),
                A.MinusOp, A.IntExp Int32.one, context 1 1), context 1 2) }
| ID MINUSMINUS
  { A.AssignStmt(A.SimpleVar($1, context 1 1),
        A.OpExp(A.VarExp(A.SimpleVar($1, context 1 1)),
                A.MinusOp, A.IntExp Int32.one, context 1 1), context 1 2) }
| BREAK
  { A.BreakStmt(context 1 1) }
| CONTINUE
  { A.ContinueStmt(context 1 1) }
| WRITEINT intexp
  { A.WriteintStmt($2, context 1 1) }

control :
  IF LPAREN boolexp RPAREN block
  { A.IfStmt($3, $5, A.SeqStmt([]), context 1 5) }
| IF LPAREN boolexp RPAREN block ELSE block
  { A.IfStmt($3, $5, $7, context 1 7) }
| WHILE LPAREN boolexp RPAREN block
  { A.WhileStmt($3, $5, context 1 5) }
| FOR LPAREN ID ASSIGN intexp SEMICOL boolexp SEMICOL simps RPAREN block
  { A.ForStmt(A.SimpleVar($3, context 1 1), $5, $7, A.SeqStmt($9), $11, context 1 11) }

block :
  stmt
  { $1 }
| LBRACE stmtseq RBRACE
  { A.SeqStmt $2 }

intexp :
  INTCONST
  { A.IntExp($1) }
| ID
  { A.VarExp(A.SimpleVar($1, context 1 1)) }
| incdec
  { $1 }
| NEG intexp
  { A.OpExp(A.IntExp Int32.zero, A.MinusOp, $2, context 1 2) }
| intexp DIVIDE intexp
  { A.OpExp($1, A.DivideOp, $3, context 1 3) }
| intexp TIMES intexp
  { A.OpExp($1, A.TimesOp, $3, context 1 3) }
| intexp PLUS intexp
  { A.OpExp($1, A.PlusOp, $3, context 1 3) }
| intexp MINUS intexp
  { A.OpExp($1, A.MinusOp, $3, context 1 3) }
| intexp MOD intexp
  { A.OpExp($1, A.ModOp, $3, context 1 3) }
| LPAREN intexp RPAREN
  { $2 }

incdec :
  PLUSPLUS ID
  { A.PreIncExp(A.SimpleVar($2, context 2 2), context 1 2) }
| ID PLUSPLUS
  { A.PostIncExp(A.SimpleVar($1, context 1 1), context 1 2) }
| MINUSMINUS ID
  { A.PreDecExp(A.SimpleVar($2, context 2 2), context 1 2) }
| ID MINUSMINUS
  { A.PostDecExp(A.SimpleVar($1, context 1 1), context 1 2) }


boolexp :
  TRUE
  { A.True(context 1 1) }
| FALSE
  { A.False(context 1 1) }
| NOT boolexp
  { A.Unopb($2, A.NotOp, context 1 2) }
| boolexp AND boolexp
  { A.Binopb($1, A.AndOp, $3, context 1 3) } 
| boolexp OR boolexp
  { A.Binopb($1, A.OrOp, $3, context 1 3) }
| boolexp XOR boolexp
  { A.Binopb($1, A.XorOp, $3, context 1 3) }
| intexp cmpop intexp
  { A.Cmpop($1, $2, $3, context 1 3) }
| LPAREN boolexp RPAREN
  { $2 }

cmpop :
  EQ
  { A.EqOp }
| NE
  { A.NeOp }
| LT
  { A.LtOp }
| LE
  { A.LeOp }
| GT
  { A.GtOp }
| GE
  { A.GeOp }

%%
